/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

/**
 * <p>Class {@code BinnedIbd} stores a list of IBD segment lengths and
 * the number of IBD segments with 0, 1, and 2 censored end-points having
 * each of these segment lengths.
 * </p>
 * Instances of class {@code BinnedIbd} are immutable.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class BinnedIbd {

    private final long nHapPairs;
    private final double[] regionLengths;
    private final double minCm;
    private final int[][] ibdCnts;
    private final double[] ibdLengths;


    /**
     * Constructs a new {@code BinnedIbd} instance.
     * @param ibd the IBD segments.
     * @throws NullPointerException if {@code ibd==null}
     */
    public BinnedIbd(Ibd ibd) {
        this.nHapPairs = ibd.nHapPairs();
        this.regionLengths = ibd.trimmedRegionLengths();
        this.minCm = ibd.par().mincm();
        this.ibdCnts = ibd.binCnts();
        this.ibdLengths = ibd.binLengths();
    }

    /**
     * Returns the number of haplotype pairs.  The returned value
     * does not include haplotype pairs that have been excluded.
     * @return the number of haplotype pairs.
     */
    public long nHapPairs() {
        return nHapPairs;
    }

    /**
     * Returns the number of continuous genomic regions.
     * @return the number of continuous genomic regions.
     */
    public int nRegions() {
        return regionLengths.length;
    }

    /**
     * Returns the minimum threshold on IBD segment length.
     * @return the minimum threshold on IBD segment length.
     */
    public double minCm() {
        return minCm;
    }

    /**
     * Returns the length of the specified genomic region.
     * @param index the index of a genomic region.
     * @return the length of the specified genomic region.
     * @throws IndexOutOfBoundsException if
     * {@code index<0 || index>= this.nRegions()}
     */
    public double regionLength(int index) {
        return regionLengths[index];
    }

    /**
     * Returns the number of IBD segment lengths that are represented by
     * {@code this}.
     * @return the number of IBD segment lengths that are represented by
     * {@code this}.
     */
    public int nIbdLengths() {
        return ibdLengths.length;
    }

    /**
     * Returns the specified IBD segment length.
     * @param index a length index.
     * @return the specified IBD segment length.
     * @throws IllegalArgumentException if
     * {@code index<0 || index>=this.nIbdLengths()}
     */
    public double ibdLength(int index) {
        return ibdLengths[index];
    }

    /**
     * Returns the number of IBD segments having length
     * {@code this.ibdLength(lengthIndex)} and having the specified number
     * of censored ends.
     * @param nCensoredEnds the number of censored ends.
     * @param lengthIndex the index of the IBD segment length.
     * @return the number of IBD segments having length
     * {@code this.ibdLength(lengthIndex)} and having the specified number
     * of censored ends.
     * @throws IndexOutOfBoundsException if
     * {@code nCensoredEnds<0 || nCensoredEnds>2}
     * @throws IndexOutOfBoundsException if
     * {@code lengthIndex<0 || lengthIndex>=this.nIbdLengths()}
     */
    public int ibdCnt(int nCensoredEnds, int lengthIndex) {
        return ibdCnts[nCensoredEnds][lengthIndex];
    }
}
