/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import beagleutil.ChromIds;
import blbutil.Const;
import blbutil.FileUtil;
import java.io.File;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import vcf.GeneticMap;

/**
 * <p>Class {@code ExtremeRegionFilter} represents a filter that excludes
 * genomic regions with extremely low or extremely high numbers of IBD segments.
 * </p>
 * <p>Instances of class {@code ExtremeRegionFilter} are immutable.
 * </p>
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class ExtremeRegionFilter {

    private final NePar par;

    /**
     * Constructs a new {@code ExtremeRegionFilter} instance.
     * @param par the analysis parameters.
     *
     * @throws IllegalArgumentException if {@code Double.isNaN(par.interval())
     *      || Double.isInfinite(par.interval()) || par.interval()<=0.0}
     * @throws IllegalArgumentException if {@code Double.isNaN(par.trimprop())
     *      || par.trimprop()<0.0 || par.trimprop()>=0.50}
     * @throws NullPointerException if {@code par==null}
     */
    public ExtremeRegionFilter(NePar par) {
        if (Double.isNaN(par.interval()) || Double.isInfinite(par.interval())
                || par.interval()<=0.0) {
            throw new IllegalArgumentException(String.valueOf(par.interval()));
        }
        if (Double.isNaN(par.trimprop()) || par.trimprop()<0.0
                || par.trimprop()>=0.50) {
            throw new IllegalArgumentException(String.valueOf(par.trimprop()));
        }
        this.par = par;
    }

    /**
     * Returns the analysis parameters.
     * @return the analysis parameters.
     */
    public NePar par() {
        return par;
    }

    /**
     * Modifies {@code ibdLists} to excludes genomic regions that have
     * an extremely low or an extremely high number of IBD segments.
     * <p>
     * First, the number of IBD segments at overlapping each locus is counted.
     * A locus is considered to have an extreme number of IBD segments if the
     * number of IBD segments is more than {@code this.par().maxdev()} trimmed
     * standard deviations away from the trimmed mean, with the trim proportion
     * equal to {@code this.par().trimprop()}. Boundaries of IBD
     * segments that intersect excluded regions are modified so that the
     * segments do not intersect excluded regions. After these modifications,
     * any IBD segment with length less than {@code this.par().minibd()}
     * centiMorgans is excluded.  After excluding loci with an extreme
     * number of IBD segment, any continuous genomic region that is less than
     * {@code this.par().minregion()} centiMorgans in length is excluded.
     * </p>
     * @param ibdLists the list of continuous regions with IBD segments.
     * @param genMap the genetic map.
     * @throws NullPointerException if {@code ibdLists==null || genMap==null}
     * @throws NullPointerException if any element of {@code ibdLists} is
     * {@code null}
     */
    public void removeExtemeRegions(List<IbdList> ibdLists, GeneticMap genMap) {
        long[] counts = counts(ibdLists, par.interval());
        counts = sortAndTrim(counts, par.trimprop());
        double mean = mean(counts);
        double stdDev = stdDev(counts, mean);
        long minCnt = (long) Math.round(mean - par.maxdev()*stdDev);
        long maxCnt = (long) Math.round(mean + par.maxdev()*stdDev);

        List<IbdList> newList = new ArrayList<>(3*ibdLists.size()/2);
        File exclFile = new File(par.out() + ".region.excl");
        try (PrintWriter exclOut=FileUtil.printWriter(exclFile)) {
            printExcludedRegionsHeader(mean, stdDev, minCnt, maxCnt, exclOut);
            for (IbdList ibdList : ibdLists) {
                List<IbdList> regions = removeExtremeRegions(ibdList, minCnt,
                        maxCnt);
                newList.addAll(regions);
                printExcludedRegions(ibdList, regions, genMap, exclOut);
            }
        }
        ibdLists.clear();
        ibdLists.addAll(newList);
    }

    /*
     * Returns an array of length whose elements are the number of IBD
     * segments that occur at positions that are separated by
     * the specified {@code interval}.
     */
    private static long[] counts(List<IbdList> ibdLists, double interval) {
        int sumLengths = 0;
        List<long[]> list = new ArrayList<>(ibdLists.size());
        for (IbdList ibdList : ibdLists) {
            long[] la = counts(ibdList, interval);
            list.add(la);
            sumLengths += la.length;
        }
        long[] counts = new long[sumLengths];
        int startIndex = 0;
        for (long[] la : list) {
            System.arraycopy(la, 0, counts, startIndex, la.length);
            startIndex += la.length;
        }
        assert startIndex==counts.length;
        return counts;
    }

    /**
     * Returns an array of length whose {@code k}-th element is the number
     * of IBD segments that overlap the position
     * {@code ibdList.minIbdStart() + (k+0.5)*interval}.
     */
    private static long[] counts(IbdList ibdList, double interval) {
        double minStart = ibdList.minIbdStart();
        double maxEnd = ibdList.maxIbdEnd();
        double length = maxEnd - minStart;
        long[] cnts = new long[1 + (int) Math.floor( (length/interval) - 0.5f)];
        for (int j=0, n=ibdList.size(); j<n; ++j) {
            int startIndex = startIndex(ibdList, j, interval);
            int endIndex = endIndex(ibdList, j, interval);
            for (int k=startIndex; k<=endIndex; ++k) {
                ++cnts[k];
            }
        }
        return cnts;
    }

    private static int startIndex(IbdList ibdList, int ibdIndex, double interval) {
        double offset = ibdList.start(ibdIndex) - ibdList.minIbdStart();
        return (int) Math.ceil( (offset/interval) - 0.5f);
    }

    private static int endIndex(IbdList ibdList, int ibdIndex, double interval) {
        double offset = ibdList.end(ibdIndex) - ibdList.minIbdStart();
        return (int) Math.floor( (offset/interval) - 0.5f);
    }

    /* Returns a list of retained regions */
    private List<IbdList> removeExtremeRegions(IbdList ibdList, long minCnt,
            long maxCnt) {
        double minIbdStart = ibdList.minIbdStart();
        List<IbdList> list = new ArrayList<>();
        long[] cnts = counts(ibdList, par.interval());
        int[] range = firstExcludedIndexRange(cnts, 0, minCnt, maxCnt);
        if (range[0]==range[1]) {
            double startPos =  ibdList.minIbdStart();
            double endPos = ibdList.maxIbdEnd();
            if ((endPos - startPos) >= par.minregion()) {
                list.add(ibdList);
            }
        }
        else {
            double startPos = minIbdStart;
            while (range[0]!=range[1]) {
                double endPos = minIbdStart + (range[0]+0.5f)*par.interval();
                if ((endPos - startPos) >= par.minregion()) {
                    list.add(restrict(ibdList, startPos, endPos, par.mincm()));
                }
                startPos = minIbdStart + (range[1] + 0.5f)*par.interval();
                range = firstExcludedIndexRange(cnts, range[1], minCnt, maxCnt);
            }
            double endPos = ibdList.maxIbdEnd();
            if ( (endPos - startPos) >= par.minregion()) {
                list.add(restrict(ibdList, startPos, endPos, par.mincm()));
            }
        }
        return list;
    }

    private static int[] firstExcludedIndexRange(long[] cnt, int start,
            long minCnt, long maxCnt) {
        while (start<cnt.length && inRange(cnt[start], minCnt, maxCnt)) {
            ++start;
        }
        int end = start;
        while (end<cnt.length && inRange(cnt[end], minCnt, maxCnt)==false) {
            ++end;
        }
        if (start>0 && start<end) {
            start = start-1;
        }
        return new int[] {start, end}; /* end may equal cnt.length */
    }

    private static boolean inRange(long cnt, long minCnt, long maxCnt) {
        return minCnt<=cnt && cnt <=maxCnt;
    }

    private static long[] sortAndTrim(long[] counts, double trimProp) {
        Arrays.sort(counts);
        int nTrim = (int) Math.ceil(trimProp*counts.length);
        assert nTrim >=0;
        int start = nTrim;
        int end = Math.max(nTrim, (counts.length-nTrim) );
        if (end<=start) {
            String s = "trim parameter for trimmed mean is too large";
            throw new IllegalStateException(s);
        }
        else {
            return Arrays.copyOfRange(counts, start, end);
        }
    }

    private static double mean(long[] counts) {
        assert counts.length>0;
        long n = 0;
        double sum = 0;
        for (long cnt : counts) {
            ++n;
            sum += cnt;
        }
        return sum/n;
    }

    private static double stdDev(long[] counts, double mean) {
        assert counts.length>0;
        long n = 0;
        double sumSquaredDev = 0;
        for (long cnt : counts) {
            ++n;
            double dev = (cnt-mean);
            sumSquaredDev += (dev*dev);
        }
        return Math.sqrt(sumSquaredDev/n);
    }

    /*
     * Constructs and returns a new {@code IbdList} instance obtained by
     * copying the list of IBD segments represented by {@code ibdList},
     * changing copied IBD segment starting points that are less than
     * {@code start} to {@code start}, changing copied IBD segment
     * end positions that are greater than {@code end} to {@code end},
     * and excluding any copied IBD segment that has length less than
     * {@code minLength} either before or after adjustment of the segment end
     * positions.
     */
    private static IbdList restrict(IbdList ibdList, double start, double end,
            double minLength) {
        assert start<=end;
        IbdList restricted = new IbdList(ibdList.chrom());
        for (int j=0, n=ibdList.size(); j<n; ++j) {
            int s1 = ibdList.sample1(j);
            int s2 = ibdList.sample2(j);
            double oldStart = ibdList.start(j);
            double oldEnd = ibdList.end(j);
            double newStart = (oldStart < start) ? start : oldStart;
            double newEnd = (oldEnd > end) ? end : oldEnd;
            if (newEnd - newStart >= minLength) {
                restricted.addIbdSegment(s1, s2, newStart, newEnd);
            }
        }
        return restricted;
    }

    private static void printExcludedRegionsHeader(double mean, double stdDev,
            long minCnt, long maxCnt, PrintWriter out) {
        DecimalFormat df1 = new DecimalFormat("0.#");
        out.print("##Trimmed.mean.IBD.segments.per.locus=");
        out.println(df1.format(mean));
        out.print("##Trimmed.stddev.IBD.segments.per.locus=");
        out.println(df1.format(stdDev));
        out.print("##Min.permitted.IBD.segments.per.locus=");
        out.println(minCnt<0 ? 0 : minCnt);
        out.print("##Max.permitted.IBD.segments.per.locus=");
        out.println(maxCnt);
        out.print("#CHROM");
        out.print(Const.tab);
        out.print("START.POS");
        out.print(Const.tab);
        out.println("END.POS");
    }

    private static void printExcludedRegions(IbdList orig,
            List<IbdList> regions, GeneticMap genMap, PrintWriter out) {
        String chrom = orig.chrom();
        double chromStart = orig.minIbdStart();
        double start = chromStart;
        for (IbdList region : regions) {
           double end = region.minIbdStart();
           if (start<end) {
               printInterval(chrom, chromStart, start, end, genMap, out);
           }
           start = region.maxIbdEnd();
        }
        double end = orig.maxIbdEnd();
        if (start<end) {
            printInterval(chrom, chromStart, start, end, genMap, out);
        }
    }

    private static void printInterval(String chrom, double minIbdStart,
            double start, double end, GeneticMap map, PrintWriter out) {
        int chromIndex = ChromIds.instance().getIndex(chrom);
        int startPos = map.basePos(chromIndex, start);
        int endPos = map.basePos(chromIndex, end);
        if (start==minIbdStart && startPos>=0) {
             startPos = 0;
        }
        out.print(chrom);
        out.print(Const.tab);
        out.print(startPos);
        out.print(Const.tab);
        out.println(endPos);
    }
}
