/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import java.util.Arrays;

/**
 * <p>Class {@code HarmonicMean} calculates the harmonic means at each array
 * index for a set of fixed-length {@code double[]} arrays with non-negative
 * elements.
 * </p>
 *
 * Instances of class {@code HarmonicMean} are not thread-safe.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class HarmonicMean {

    private int nArrays;
    private final double[] sumRecipricals;

    /**
     * Constructs a new instance of {@code HarmonicMean}.
     * @param length an array length
     * @throws NegativeArraySizeException if {@code length < 0}
     */
    public HarmonicMean(int length) {
        this.nArrays = 0;
        this.sumRecipricals = new double[length];
    }

    /**
     * Clears all array data.
     */
    public void clear() {
        this.nArrays = 0;
        Arrays.fill(sumRecipricals, 0.0);
    }

    /**
     * Returns the array length.
     * @return the array length
     */
    public int length() {
        return sumRecipricals.length;
    }

    /**
     * Returns the number of arrays with stored data.
     * @return the number of arrays with stored data
     */
    public int nArrays() {
        return nArrays;
    }

    /**
     * Include the elements of the specific array in the harmonic mean
     * calculation.
     * @param da an array of non-zero elements
     * @throws IllegalArgumentException if
     * {@code da.length != this.length()}
     * @throws IllegalArgumentException if
     * {@code (da[j] <= 0.0 || Double.isNaN(da[j])} for any element
     * {@code j} satisfying {@code (0 <= j && j < this.nArrays())}
     */
    public void add(double[] da) {
        if (da.length != length()) {
            throw new IllegalArgumentException(String.valueOf(da.length));
        }
        ++nArrays;
        for (int j=0; j<da.length; ++j) {
            if (da[j]<=0.0 || Double.isNaN(da[j])) {
                throw new IllegalArgumentException(String.valueOf(da[j]));
            }
            sumRecipricals[j] += 1.0/da[j];
        }
    }

    /**
     * Returns the harmonic means of the stored data at each array index.
     * @return the harmonic means of the stored data at each array index
     * @throws IllegalArgumentException if {@code this.nArrays() == 0}
     */
    public double[] mean() {
        if (nArrays==0) {
            throw new IllegalArgumentException("0");
        }
        return Arrays.stream(sumRecipricals)
                .parallel()
                .map(d -> nArrays/d)
                .toArray();
    }
}
