/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import java.util.Random;

/**
 * <p>Class {@code Ibd} contains length and end-censoring data
 * for IBD segments contained in continuous genomic regions.
 * </p>
 * Instances of class {@code Ibd} are immutable.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class Ibd {

    static final int CENSORED_LENGTH = 3;   // # of censoring categories

    private final NePar par;
    private final int nRegions;
    private final long nHapPairs;
    private final long nInputSegments;
    private final double[] regionLengths;
    private final double[] binLengths;
    private final int[][][] binCnts;   //[region][# censored ends][count]

    /**
    * <p>Constructs a new {@code Ibd} instance for the specified IBD segments
    * in each continuous genomic region.
    * </p>
    * <p>Each end of each continuous genomic region is trimmed by
    * {@code ibdReader.par().trimcm()} centiMorgans.  IBD segment endpoints
    * are censored if an IBD segment overlaps with the removed region ends.
    * IBD segments are excluded if their length after region end trimming
    * is less than {@code ibdReader.par().mincm()} centiMorgans.
    * </p>
    * @param ibdReader the IBD segments for each continuous genomic region.
    *
    * @throws IllegalArgumentException if {@code ibdReader.par().mincm()}
    * is non-positive, infinite, or {@code NaN}.
    * @throws IllegalArgumentException if {@code ibdReader.par().trimcm()}
    * is negative, infinite, or {@code NaN}.
    * @throws IllegalArgumentException if {@code ibdReader.par().binwidth()}
    * is non-positive, infinite, or {@code NaN}.
    * @throws NullPointerException if {@code ibdReader==null}
    */
    public Ibd(IbdReader ibdReader) {
        this.par = ibdReader.par();
        double minIbdLength = par.mincm();
        double regionTrim = par.trimcm();
        double binWidth = par.binwidth();
        checkParameters(minIbdLength, regionTrim, binWidth);
        double[] maxSegmentLengths = maxSegmentLengths(ibdReader);

        this.nRegions = ibdReader.nIbdLists();
        this.nHapPairs = nHapPairs(ibdReader);
        this.nInputSegments = ibdReader.nInputIbdSegments();
        this.regionLengths = regionLengths(ibdReader, regionTrim);
        this.binLengths = binCenters(minIbdLength, maxSegmentLengths, binWidth);
        this.binCnts = initBinCnts(nRegions, minIbdLength,
                maxSegmentLengths, binWidth);
        fillBinCnts(ibdReader, binCnts, regionTrim, binWidth);
    }

    private static void checkParameters(double mincm, double trimcm,
            double binWidth) {
        if (mincm <= 0.0 || Double.isInfinite(mincm) || Double.isNaN(mincm)) {
            throw new IllegalArgumentException("mincm: " + mincm);
        }
        if (trimcm < 0 || Double.isInfinite(trimcm) || Double.isNaN(trimcm)) {
            throw new IllegalArgumentException("trimcm: " + trimcm);
        }
        if (binWidth <= 0 || Double.isInfinite(binWidth)
                || Double.isNaN(binWidth)) {
            throw new IllegalArgumentException("binwidth: " + binWidth);
        }
    }

    /**
     * Returns a new {@code IBD} instance with {@code this.nRegions()}
     * continuous genomic regions that is obtained by sampling with replacement
     * the continuous genomic regions represented by {@code this}.
     * @param seed a seed for a random number generator.
     * @return a new {@code IBD} instance with {@code this.nRegions()}
     * continuous genomic regions that is obtained by sampling with replacement
     * the continuous genomic regions represented by {@code this}.
     */
    public Ibd bootstrapSample(long seed) {
        return new Ibd(this, seed);
    }

    private Ibd(Ibd ibd, long seed) {
        this.par = ibd.par;
        this.nRegions = ibd.nRegions;
        this.nHapPairs = ibd.nHapPairs;
        this.nInputSegments = ibd.nInputSegments;
        this.binLengths = ibd.binLengths;

        Random random = new Random(seed);
        this.binCnts = new int[nRegions][][];
        this.regionLengths = new double[nRegions];
        for (int i=0; i<nRegions; ++i) {
            int region = random.nextInt(nRegions);
            this.binCnts[i] = ibd.binCnts[region];
            this.regionLengths[i] = ibd.regionLengths[region];
        }
    }

    private static double[] maxSegmentLengths(IbdReader reader) {
        double[] maxIbdSegmentLengths = new double[reader.nIbdLists()];
        for (int i=0; i<maxIbdSegmentLengths.length; ++i) {
            maxIbdSegmentLengths[i] = reader.ibdList(i).maxIbdLength();
        }
        return maxIbdSegmentLengths;
    }

    private static double[] regionLengths(IbdReader reader, double endTrim) {
        double totalChromTrim = 2*endTrim;
        double[] trimmedLengths = new double[reader.nIbdLists()];
        for (int i=0; i<trimmedLengths.length; ++i) {
            IbdList ibdList = reader.ibdList(i);
            double length = ibdList.maxIbdEnd() - ibdList.minIbdStart();
            if (length > totalChromTrim) {
                trimmedLengths[i] = length - totalChromTrim;
            }
        }
        return trimmedLengths;
    }

    /*
     * Excludes IBD segments that are less than length threshold after
     * trimming chromosome ends
     */
    private static void fillBinCnts(IbdReader reader, int[][][] binCnts,
            double chromTrim, double binWidth) {
        int nIbdLists = reader.nIbdLists();
        double minIbdLength = reader.par().mincm();
        for (int i=0; i<nIbdLists; ++i) {
            IbdList ibdList = reader.ibdList(i);
            double trimmedChromStart = ibdList.minIbdStart() + chromTrim;
            double trimmedChromEnd = ibdList.maxIbdEnd() - chromTrim;
            for (int j=0, n=ibdList.size(); j<n; ++j) {
                int censoredCnt = 0;
                double start = ibdList.start(j);
                double end = ibdList.end(j);
                if (start <= trimmedChromStart) {
                    start = trimmedChromStart;
                    ++censoredCnt;
                }
                if (end >= trimmedChromEnd) {
                    end = trimmedChromEnd;
                    ++censoredCnt;
                }
                double length = (end - start);
                if (minIbdLength<=length) {
                    int binIndex = binIndex(minIbdLength, length, binWidth);
                    ++binCnts[i][censoredCnt][binIndex];
                }
            }
        }
    }

    private static int[][][] initBinCnts(int nRegions, double minIbdLength,
            double[] maxIbdLength, double binSize) {
        assert nRegions==maxIbdLength.length;
        int[][][] bc = new int[nRegions][CENSORED_LENGTH][];
        for (int i=0; i<nRegions; ++i) {
            int nBins = 0;
            if (maxIbdLength[i] > minIbdLength) {
                nBins = 1 + binIndex(minIbdLength, maxIbdLength[i], binSize);
            }
            for (int j=0; j<CENSORED_LENGTH; ++j) {
                bc[i][j] = new int[nBins];
            }
        }
        return bc;
    }

    private static double[] binCenters(double minIbdLength, double[] maxIbdLength,
            double binWidth) {
        double max = max(maxIbdLength);
        int nBins = 1 + binIndex(minIbdLength, max, binWidth);
        double[] centers = new double[nBins];
        for (int j=0; j<centers.length; ++j) {
            centers[j] = minIbdLength + (j+0.5f)*binWidth;
        }
        return centers;
    }

    private static int binIndex(double minIbdLength, double ibdLength,
            double binSize) {
        if (ibdLength < minIbdLength) {
            return -1;
        }
        else {
            return (int) Math.floor( (ibdLength - minIbdLength)/binSize);
        }
    }

    private static double max(double[] fa) {
        double max = Double.NEGATIVE_INFINITY;
        for (double d : fa) {
            if (d > max) {
                max = d;
            }
        }
        return max;
    }

    private static long nHapPairs(IbdReader ibdReader) {
        long nPairs = ibdReader.par().npairs();
        if (nPairs>0) {
            return nPairs;
        }
        else {
            long n = 2*ibdReader.nSamples();
            return ( (n*(n-2)/2) - 4*ibdReader.nExcludedSamplePairs() );
        }
    }

    /**
     * Returns the analysis parameters.
     * @return the analysis parameters.
     */
    public NePar par() {
        return par;
    }

    /**
     * Returns the number of continuous genomic regions.
     * @return the number of continuous genomic regions.
     */
    public int nRegions() {
        return nRegions;
    }

    /**
     * Returns an array of length {@code this.nRegions()}, whose {@code j}-th
     * element is the length of the {@code j}-th trimmed, continuous genomic
     * region.
     *
     * @return an array of length {@code this.nRegions()}, whose {@code j}-th
     * element is the length of the {@code j}-th trimmed, continuous genomic
     * region.
     */
    public double[] trimmedRegionLengths() {
        return regionLengths.clone();
    }

    /**
     * Returns the number of unordered haplotype pairs after excluding
     * haplotype pairs from closely related samples and haplotype pairs
     * from the same sample.
     * @return the number of unordered haplotype pairs after excluding
     * haplotype pairs from closely related samples and haplotype pairs
     * from the same sample.
     */
    public long nHapPairs() {
        return nHapPairs;
    }

    /**
     * Returns a two-dimensional array whose {@code [j][k]}-th element
     * is the number or IBD segments with {@code j} censored
     * ends {@code (0<=j<2)}, length greater than or equal to
     * {@code (this.par().mincm() + k*this.par().binWidth())}, and length
     * less than {@code (this.par().mincm() + (k+1)*this.binWidth())}.
     *
     * @return a two dimensional array with IBD counts for each bin
     * of IBD segment length and number of censored end coordinates.
     */
    public int[][] binCnts() {
        int[][] sumBinCnts = new int[CENSORED_LENGTH][binLengths.length];
        for (int i=0; i<nRegions; ++i) {
            for (int j=0; j<CENSORED_LENGTH; ++j) {
                for (int k=0; k<binCnts[i][j].length; ++k) {
                    sumBinCnts[j][k] += binCnts[i][j][k];
                }
            }
        }
        return sumBinCnts;
    }

    /**
     * Returns an array whose {@code k}-th element is
     * {@code (this.par().mincm() + (k+0.5)*this.par.binWidth())}, which
     * is the mean of the range of IBD segments lengths represented
     * by the {@code k}-th bin.
     *
     * @return an array whose {@code k}-th element is the mean of the
     * range of IBD segments lengths represented by the {@code k}-th bin.
     */
    public double[] binLengths() {
        return binLengths.clone();
    }

    /**
     * Returns the number of input IBD segments.
     * @return the number of input IBD segments.
     */
    public long nInputSegments() {
        return nInputSegments;
    }

    /**
     * Returns the number of IBD segments included in the analysis.
     * @return the number of IBD segments included in the analysis.
     */
    public long nSegmentsAnalyzed() {
        long sum = 0;
        for (int[][] i3 : binCnts) {
            for (int[] i2 :  i3) {
                for (int i1 : i2) {
                    if (i1>0) {
                        sum += i1;
                    }
                }
            }
        }
        return sum;
    }
}
