/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import java.util.Arrays;

/**
 * Class {@code IbdList} stores a list of IBD segments on a chromosome.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public final class IbdList {
    private static final int INIT_NSEGS = 10000;

    private final String chrom;
    private int size = 0;
    private int[] id1 = new int[INIT_NSEGS];
    private int[] id2 = new int[INIT_NSEGS];
    private double[] startPos = new double[INIT_NSEGS];
    private double[] endPos = new double[INIT_NSEGS];

    private double maxIbdLength = 0f;
    private double minIbdStart = Double.MAX_VALUE;
    private double maxIbdEnd = -Double.MAX_VALUE;

    /**
     * Constructs a new {@code IbdList} instance.
     *
     * @param chrom the chromosome.
     * @throws NullPointerException if {@code chrom==null}
     */
    public IbdList(String chrom) {
        if (chrom==null) {
            throw new NullPointerException(chrom);
        }
        this.chrom = chrom;
    }

    /**
     * Returns the chromosome.
     * @return the chromosome.
     */
    public String chrom() {
        return chrom;
    }

    /**
     * Adds the specified IBD segment to this list.
     * @param sample1 the first sample index.
     * @param sample2 the second sample index.
     * @param start the starting position of the IBD segment.
     * @param end the ending position of the IBD segment.
     * @throws IllegalArgumentException if {@code sample1>sample2}
     * @throws IllegalArgumentException if {@code start>=end}
     * @throws IllegalArgumentException if
     * {@code Double.isNaN(start) || Double.isInfinite(start)}
     * @throws IllegalArgumentException if
     * {@code Double.isNaN(end) || Double.isInfinite(end)}
     */
    public void addIbdSegment(int sample1, int sample2, double start, double end) {
        checkSamples(sample1, sample2);
        checkInterval(start, end);
        ensureCapacity();
        id1[size] = sample1;
        id2[size] = sample2;
        startPos[size] = (double) start;
        endPos[size] = (double) end;
        ++size;
        if (start < minIbdStart) {
            minIbdStart = start;
        }
        if (end > maxIbdEnd) {
            maxIbdEnd = end;
        }
        double length = end - start;
        if (length > maxIbdLength) {
            maxIbdLength = length;
        }
    }

    private void ensureCapacity() {
        if (size==id1.length) {
            int newCapacity = 3*size/2 + 1;
            id1 = Arrays.copyOf(id1, newCapacity);
            id2 = Arrays.copyOf(id2, newCapacity);
            startPos = Arrays.copyOf(startPos, newCapacity);
            endPos = Arrays.copyOf(endPos, newCapacity);
        }
    }

    private static void checkSamples(int sample1, int sample2) {
        if (sample1>sample2) {
            String s = "sample1=" + sample1 + " sample2=" + sample2;
            throw new IllegalArgumentException(s);
        }
    }

    private static void checkInterval(double start, double end) {
        if (start>=end) {
            String s = "start=" + start + " end=" + end;
            throw new IllegalArgumentException(s);
        }
        checkDouble(start);
        checkDouble(end);
    }

    private static void checkDouble(double d) {
        if (Double.isNaN(d) || Double.isInfinite(d)) {
            throw new IllegalArgumentException(String.valueOf(d));
        }
    }

   /**
     * Returns the minimum starting position for the IBD
     * segments represented by {@code this}.
     * Returns 0 if {@code this.size()==0}.
     * @return the minimum starting position for the IBD
     * segments represented by {@code this}.
     */
    public double minIbdStart() {
        return size==0 ? 0 : minIbdStart;
    }

   /**
     * Returns the maximum ending position for the IBD
     * segments represented by {@code this}.
     * Returns 0 if {@code this.size()==0}.
     * @return the maximum ending position for the IBD
     * segments represented by {@code this}.
     */
    public double maxIbdEnd() {
        return size==0 ? 0 : maxIbdEnd;
    }

    /**
     * Returns the maximum IBD segment length for the IBD segments
     * represented by {@code this}.
     * @return the maximum IBD segment length for the IBD segments
     * represented by {@code this}.
     */
    public double maxIbdLength() {
        return maxIbdLength;
    }

   /**
     * Returns the number of IBD segments represented by {@code this}.
     * @return the number of IBD segments represented by {@code this}.
     */
    public int size() {
        return size;
    }

    /**
     * Returns the first sample index of the specified IBD segment.
     * @param index an IBD segment index.
     * @return the first sample index of the specified IBD segment.
     * @throws IndexOutOfBoundsException if
     * {@code index<0 || index>=this.size()}
     */
    public int sample1(int index) {
        if (index>=size) {
            throw new IndexOutOfBoundsException(String.valueOf(index));
        }
        return id1[index];
    }

    /**
     * Returns the second sample index of the specified IBD segment.
     * @param index an IBD segment index.
     * @return the second sample index of the specified IBD segment.
     * @throws IndexOutOfBoundsException if
     * {@code index<0 || index>=this.size()}
     */
    public int sample2(int index) {
        if (index>=size) {
            throw new IndexOutOfBoundsException(String.valueOf(index));
        }
        return id2[index];
    }

    /**
     * Returns the starting position of the specified IBD segment.
     * @param index an IBD segment index.
     * @return the starting position of the specified IBD segment.
     * @throws IndexOutOfBoundsException if
     * {@code index<0 || index>=this.size()}
     */
    public double start(int index) {
        if (index>=size) {
            throw new IndexOutOfBoundsException(String.valueOf(index));
        }
        return startPos[index];
    }

    /**
     * Returns the ending position of the specified IBD segment.
     * @param index an IBD segment index.
     * @return the ending position of the specified IBD segment.
     * @throws IndexOutOfBoundsException if
     * {@code index<0 || index>=this.size()}
     */
    public double end(int index) {
        if (index>=size) {
            throw new IndexOutOfBoundsException(String.valueOf(index));
        }
        return endPos[index];
    }

    /**
     * Returns a string representation of {@code this}. The exact details
     * of the representation are unspecified and subject to change.
     * @return a string representation of {@code this}.
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("chrom=");
        sb.append(chrom);
        sb.append("  minIbdStart=");
        sb.append(minIbdStart);
        sb.append("  maxIbdEnd=");
        sb.append(maxIbdEnd);
        sb.append("  size=");
        sb.append(size);
        return sb.toString();
    }
}
