/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import beagleutil.SampleIds;
import blbutil.Const;
import blbutil.FileIt;
import blbutil.FileUtil;
import blbutil.InputIt;
import blbutil.Utilities;
import ibdutil.StringIdSegment;
import java.io.File;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import vcf.GeneticMap;
import vcf.PlinkGenMap;

/**
 * <p>Class {@code IbdReader} reads, filters, and stores IBD segments
 * for each continuous genomic region.
 * See the class constructor documentation for a description of the filters
 * that are applied.
 * </p>
 * <p>Instances of class {@code IbdReader} are immutable.
 * </p>
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public final class IbdReader {

    private static final DecimalFormat df2 = new DecimalFormat("0.##");

    private final NePar par;
    private final List<IbdList> ibdLists;
    private final long nIbdSegmentsRead;
    private final int nSamples;
    private final int nExcludedPairs;

    /**
     * Filename extension for the output file containing sample
     * pairs whose IBD segments are excluded from the analysis.
     */
    public static final String PAIR_EXCL_FILE_EXT = ".pair.excl";

    /**
     * <p>Reads IBD segments from standard input, filters the IBD segments,
     * writes a list of excluded sample pairs (if any) to the file whose
     * pathname is the string
     * {@code 'par.out() + IbdReader.PAIR_EXCL_FILE_EXT'}, and
     * constructs a new {@code IbdReader} instance.
     * </p>
     * <p>Three IBD segments filters are applied in the following order:
     * </p>
     * <p>First, all IBD segments with length less than {@code par.minibd()}
     * centiMorgans are excluded.
     * </p>
     * <p>Second, the number of the remaining IBD segments at each locus is
     * counted. Any genomic region that is less than {@code par.minregion()}
     * centiMorgans in length and bounded on each side by a chromosome boundary
     * or a locus with an extreme number of IBD segments is excluded.
     * A locus is considered to have an extreme number of IBD segments if the
     * number of IBD segments is more than {@code par.maxdev()} trimmed
     * standard deviations away from the trimmed mean. Boundaries of IBD
     * segments that intersect excluded regions are modified so that the
     * segments do not intersect excluded regions. After these modifications,
     * any IBD segment with length less than {@code par.minibd()} centiMorgans
     * is excluded.
     * </p>
     * <p>Third, if {@code par.filtersamples()==true} the lengths of the
     * remaining IBD segments are summed for each sample pair, and all
     * IBD segments for a sample pair are excluded if the sum of IBD segments
     * lengths for the sample pair is greater than the product of
     * {@code par.maxsharing()} and the maximum possible sum of IBD segment
     * lengths. The maximum possible sum of IBD segment lengths is the sum
     * of the distance between the minimum IBD segment starting position and
     * the maximum IBD segment ending position in each non-excluded region.
     * The maximum possible sum of IBD segment lengths is determined after the
     * first two IBD segment filters are applied.  The third filter is not
     * applied if {@code par.filtersamples()==false}.
     * </p>
     * @param par the analysis parameters.
     * @throws IllegalArgumentException if {@code Double.isNaN(par.interval())
     *      || Double.isInfinite(par.interval()) || par.interval()<=0.0}
     * @throws IllegalArgumentException if {@code Double.isNaN(par.trimprop())
     *      || par.trimprop()<0.0 || par.trimprop()>=0.50}
     * @throws NullPointerException if {@code par==null}
     */
    public IbdReader(NePar par) {
        BitSet sampleIndices = new BitSet();
        this.par = par;
        this.ibdLists = new ArrayList<>();
        this.nIbdSegmentsRead = readIbdSegments(par, sampleIndices, ibdLists);
        this.nSamples = sampleIndices.cardinality();
        this.nExcludedPairs = excludeCloseRelatives(par, nSamples, ibdLists);
        removeEmptyLists(ibdLists);
        if (ibdLists.isEmpty()) {
            printNoIbdSegmentsErrorAndExit(par);
        }
    }

    private static void printNoIbdSegmentsErrorAndExit(NePar par) {
        StringBuilder sb = new StringBuilder(200);
        sb.append(Const.nl);
        sb.append(Const.nl);
        sb.append("ERROR: No IBD segments remain after applying filters: ");
        sb.append(Const.nl);
        sb.append("  mincm=");
        sb.append(df2.format(par.mincm()));
        sb.append(Const.nl);
        if (par.filtersamples()) {
            sb.append("  filtersamples=true");
            sb.append(Const.nl);
        }
        sb.append("--- exiting program ---");
        sb.append(Const.nl);
        Utilities.exit(sb.toString());
    }

    private static void printBootstrapErrAndExit(NePar par) {
        StringBuilder sb = new StringBuilder(200);
        sb.append(Const.nl);
        sb.append(Const.nl);
        sb.append("ERROR: Cannot bootstrap sample from a single chromosomes.");
        sb.append(Const.nl);
        sb.append("       Rerun the program with \"nboots=0\".");
        sb.append(Const.nl);
        sb.append("--- exiting program ---");
        sb.append(Const.nl);
        Utilities.exit(sb.toString());
    }

    /**
     * Returns the analysis parameters
     *
     * @return the analysis parameters.
     */
    public NePar par() {
        return par;
    }

    /**
     * Returns the number of IBD lists.  Each IBD list contains
     * the IBD segments (after filtering) in a continuous region that
     * is bounded on each side by an excluded region or a chromosome end.
     * @return the number of IBD lists.
     */
    public int nIbdLists() {
        return ibdLists.size();
    }

    /**
     * Returns the specified IBD list.  Each IBD list contains
     * the IBD segments (after filtering) in a continuous region that
     * is bounded on each side by an excluded region or a chromosome end.
     * @param index the index of the IBD list.
     * @return the specified IBD list.
     * @throws IndexOutOfBoundsException if
     * {@code index<0 || index>=this.nIbdLists()}
     */
    public IbdList ibdList(int index) {
        return ibdLists.get(index);
    }

    /**
     * Returns the number of IBD segments in the input data (before filtering).
     * @return the number of IBD segments in the input data (before filtering).
     */
    public long nInputIbdSegments() {
        return nIbdSegmentsRead;
    }

    /**
     * Returns the number of samples in the input IBD segment data.
     * @return the number of samples in the input IBD segment data.
     */
    public int nSamples() {
        return nSamples;
    }

    /**
     * Returns the number of excluded sample pairs. All IBD segments
     * for each excluded sample pair are excluded from the analysis.  A list
     * of excluded sample pairs is written to the file whose pathname is the
     * string {@code 'par.out() + IbdReader.PAIR_EXCL_FILE_EXT'}.
     *
     * @return the number of excluded sample pairs.
     */
    public long nExcludedSamplePairs() {
        return nExcludedPairs;
    }

    /* returns the number of IBD segments read */
    private static long readIbdSegments(NePar par, BitSet sampleIndices,
            List<IbdList> ibdLists) {
        GeneticMap genMap = PlinkGenMap.fromPlinkMapFile(par.map());
        double minIbdLength = par.mincm();
        long readCnt = 0;
        Map<String, IbdList> ibdMap = new HashMap<>();
        try (FileIt<String> it = InputIt.fromStdIn()) {
            while (it.hasNext()) {
                String line = it.next().trim();
                if (line.length()>0) {
                    ++readCnt;
                    StringIdSegment ibdSeg = StringIdSegment.parse(line);
                    sampleIndices.set(ibdSeg.sampleIndex1());
                    sampleIndices.set(ibdSeg.sampleIndex2());
                    IbdList ibdList = getIbdList(ibdMap, ibdSeg);
                    addIbdSegment(ibdSeg, ibdList, genMap, minIbdLength);
                }
            }
        }
        for (String chrom : ibdMap.keySet()) {
            ibdLists.add(ibdMap.get(chrom));
        }
        removeEmptyLists(ibdLists);
        if (ibdLists.isEmpty()) {
            printNoIbdSegmentsErrorAndExit(par);
        }
        if (ibdLists.size()==1 && par.nboots()!=0) {
            printBootstrapErrAndExit(par);
        }
        ExtremeRegionFilter ilh = new ExtremeRegionFilter(par);
        ilh.removeExtemeRegions(ibdLists, genMap);
        return readCnt;
    }

    private static IbdList getIbdList(Map<String, IbdList> ibdMap,
            StringIdSegment ibdSeg) {
        String chrom = ibdSeg.chrom();
        IbdList ibdList = ibdMap.get(chrom);
        if (ibdList==null) {
            ibdList = new IbdList(chrom);
            ibdMap.put(chrom, ibdList);
        }
        return ibdList;
    }

    private static void addIbdSegment(StringIdSegment ibdSeg, IbdList ibdList,
            GeneticMap map, double minIbdLength) {
        int chrIndex = ibdSeg.chromIndex();
        int s1 = ibdSeg.sampleIndex1();
        int s2 = ibdSeg.sampleIndex2();
        if (s1>s2) {
            int tmp = s1;
            s1 = s2;
            s2 = tmp;
        }
        double start = map.genPos(chrIndex, ibdSeg.start());
        double end = map.genPos(chrIndex, ibdSeg.inclEnd());
        double length = end - start;
        if (length>=minIbdLength && s1!=s2) {
            ibdList.addIbdSegment(s1, s2, start, end);
        }
    }

    /*
    * Returns the number of excluded pairs of close relative.
    */
    private static int excludeCloseRelatives(NePar par, int nSamples,
            List<IbdList> ibdLists) {
        int nExclPairs = 0;
        if (par.filtersamples()) {
            double genomeLength = genomeLength(ibdLists);
            final double maxSumIbd = par.maxsharing() * genomeLength;
            final double[][] sum = sumIbdPerPair(nSamples, ibdLists);

            // count and write excluded pairs to a file
            File exclFile = new File(par.out() + PAIR_EXCL_FILE_EXT);
            try (PrintWriter exclOut=FileUtil.printWriter(exclFile)) {
                printSampleSharingHeader(maxSumIbd, genomeLength, exclOut);
                for (int j=1; j<nSamples; ++j) {
                    for (int i=0; i<j; ++i) {
                        if (sum[j][i] > maxSumIbd) {
                            ++nExclPairs;
                            printSampleSharing(i, j, sum[j][i], exclOut);
                        }
                    }
                }
            }

            // remove IBD segments for excluded sample pairs
            excludeSamplePairs(ibdLists, sum, maxSumIbd);
        }
        return nExclPairs;
    }

    private static double[][] sumIbdPerPair(int nSamples, List<IbdList> ibdLists) {
        final double[][] sum = new double[nSamples][];
        for (int i=0; i<nSamples; ++i) {
            sum[i] = new double[i];    // pair (i,j) stored in [j][i]-th element
        }
        // sum IBD segment lengths for each sample pair
        for (IbdList ci : ibdLists) {
            if (ci!=null) {
                for (int j=0; j<ci.size(); ++j) {
                    sum[ci.sample2(j)][ci.sample1(j)] += (ci.end(j) - ci.start(j));
                }
            }
        }
        return sum;
    }

    private static void excludeSamplePairs(List<IbdList> list, double[][] sum,
            double maxIbd) {
        SamplePairFilter filter = samplePairFilter(sum, maxIbd);
        for (int j=0, n=list.size(); j<n; ++j) {
            IbdList restricted = restrict(list.get(j), filter);
            list.set(j, restricted);
        }
    }

    private static SamplePairFilter samplePairFilter(final double[][] sumIbd,
            final double maxIbd) {
        return new SamplePairFilter() {
            @Override
            public boolean accept(int sample1, int sample2) {
                if (sample1>sample2) {
                    throw new IllegalArgumentException("sample1>sample2");
                }
                if (sample1<0) {
                    throw new IllegalArgumentException("sample1<0: " + sample1);
                }
                return sumIbd[sample2][sample1] <= maxIbd;
            }
        };
    }

    private static IbdList restrict(IbdList ibdList, SamplePairFilter filter) {
        IbdList restricted = new IbdList(ibdList.chrom());
        for (int j=0, n=ibdList.size(); j<n; ++j) {
            int s1 = ibdList.sample1(j);
            int s2 = ibdList.sample2(j);
            if (filter.accept(s1, s2)) {
                double start = ibdList.start(j);
                double end = ibdList.end(j);
                restricted.addIbdSegment(s1, s2, start, end);
            }
        }
        return restricted;
    }

    private static void printSampleSharingHeader(double maxIbd,
            double genomeLength, PrintWriter exclOut) {
        exclOut.print("##Genome.length=");
        exclOut.println(df2.format(genomeLength));
        exclOut.print("##Max.permitted.total.IBD=");
        exclOut.println(df2.format(maxIbd));
        exclOut.print("#ID1");
        exclOut.print(Const.tab);
        exclOut.print("ID2");
        exclOut.print(Const.tab);
        exclOut.println("Total.IBD");
    }

    private static void printSampleSharing(int s1, int s2, double sumIbdLength,
            PrintWriter relOut) {
        relOut.print(SampleIds.instance().id(s1));
        relOut.print(Const.tab);
        relOut.print(SampleIds.instance().id(s2));
        relOut.print(Const.tab);
        relOut.println(df2.format(sumIbdLength));
    }

    private static double genomeLength(List<IbdList> ibdLists) {
        double sum = 0.0f;
        for (IbdList ibdList : ibdLists) {
            sum += (ibdList.maxIbdEnd() - ibdList.minIbdStart());
        }
        return sum;
    }

    private static void removeEmptyLists(List<IbdList> ibdLists) {
        Iterator<IbdList> it = ibdLists.iterator();
        while (it.hasNext()) {
            IbdList ibdList = it.next();
            if (ibdList.size()==0) {
                it.remove();
            }
        }
    }
}
