/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import blbutil.Const;
import java.util.stream.IntStream;

/**
 * <p>Class {@code Ne} stores estimated effective population sizes,
 * and the the observed and expected length of the identity by descent
 * segments having a common ancestor a specified number of generations
 * before the present.
 * </p>
 * Instances of class {@code Ne} are immutable.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class Ne {

    private final int gMin;
    private final double[] ne;
    private final double[] probTMRCA;
    private final BinnedIbd binnedIbd;
    private final double[] exp;
    private final double[] obs;
    private final double fit;

    /**
     * Constructs a new {@code ProbG} object for the specified historical
     * effective population sizes, generations, and observed IBD segments.
     * @param ne an array whose {@code k}-th element is the
     * effective population size {@code k} generations before
     * the present generation.
     * @param gMin the first generation before the present for which
     * results will be calculated.
     * @param binnedIbd the observed IBD segments, binned by length and number
     * of censored ends.
     *
     * @throws IllegalArgumentException if any element of {@code ne} is less
     * than 0.5, is infinite, or is {@code NaN}.
     * @throws IndexOutOfBoundsException if {@code gMin < 0 || gMin >= n.length}.
     * @throws NullPointerException if {@code n == null || binnedIbd == null}
     */
    public Ne(double[] ne, int gMin, BinnedIbd binnedIbd) {
        checkNe(ne, gMin);
        this.gMin = gMin;
        this.ne = ne.clone();
        this.probTMRCA = probTMRCA(ne, gMin);
        this.binnedIbd = binnedIbd;
        this.exp = expIbd(ne, gMin, binnedIbd, probTMRCA);
        this.obs = obsIbd(ne, gMin, binnedIbd, probTMRCA);
        this.fit = IntStream.range(gMin, ne.length)
                .parallel()
                .mapToDouble(g -> Math.pow((exp[g]-obs[g]), 2.0)/exp[g])
                .sum();
    }

    private void checkNe(double[] ne, int gMin) {
        if (gMin>=ne.length) {
            throw new IndexOutOfBoundsException(String.valueOf(gMin));
        }
        for (int g=gMin; g<ne.length; ++g) {
            double d = ne[g];
            if ((d<=0.5) || Double.isNaN(d) || Double.isInfinite(d)) {
                String s = "ERROR: Invalid estimated population size: "
                        + " ne[" + g + "]=" + ne[g] + Const.nl
                        + "There may be a problem with the input data.";
                throw new IllegalArgumentException(s);
            }
        }
    }

    private static double[] expIbd(double[] ne, int gMin, BinnedIbd binnedIbd,
            double[] probTMRCA) {
        double[] exp = new double[ne.length];
        for (int g=gMin; g<ne.length; ++g) {
            if (g<gMin) {
                throw new IndexOutOfBoundsException(String.valueOf(g));
            }
            double c = binnedIbd.minCm();
            double cgDiv50 = c*g/50.0;
            double expC = Math.exp(-cgDiv50);
            double sumC = 0.0;
            int nRegions = binnedIbd.nRegions();
            for (int j=0; j<nRegions; ++j) {
                double length = binnedIbd.regionLength(j);
                if (length>=c) {
                    sumC += ( (cgDiv50+1)*length - c*cgDiv50 );
                }
            }
            exp[g] = binnedIbd.nHapPairs()*probTMRCA[g]*expC*sumC;
        }
        return exp;
    }

    private static double[] obsIbd(double[] ne, int gMin, BinnedIbd binnedIbd,
            double[] probTMRCA) {
        double tailN = ne[ne.length-1];
        double[][][] probs = probs(probTMRCA, gMin, tailN, binnedIbd);
        double[] obs = new double[ne.length];
        for (int g=gMin; g<ne.length; ++g) {
            double ibd = 0;
            int nIbdLengths = binnedIbd.nIbdLengths();
            for (int nCens=0; nCens<Ibd.CENSORED_LENGTH; ++nCens) {
                for (int len=0; len<nIbdLengths; ++len) {
                    double length = binnedIbd.ibdLength(len);
                    int cnt = binnedIbd.ibdCnt(nCens, len);
                    double p = probs[nCens][g][len];
                    ibd += length*cnt*p;
                }
            }
            obs[g] = ibd;
        }
        return obs;
    }

    /**
     * Returns the first generation before the present for which
     * lengths of observed and expected IBD will be calculated.
     * @return the first generation before the present for which
     * lengths of observed and expected IBD will be calculated
     */
    public int gMin() {
        return gMin;
    }

    /**
     * Returns the number of generations (starting at 0) that are
     * represented by {@code this}.
     * @return the number of generations (starting at 0) that are
     * represented by {@code this}
     */
    public int nGen() {
        return ne.length;
    }

    /**
     * Returns the effective population size {@code g} generations before
     * the present. If {@code g < this.gMin()} a value of 0.0 is returned.
     * @param g the number of generations before the present.
     * @return the effective population size {@code g} generations before
     * the present.
     * @throws IndexOutOfBoundsException if {@code g < 0 || g >= this.nGen()}
     */
    public double ne(int g) {
        return ne[g];
    }

    /**
     * Returns the array of effective population sizes whose {@code g}-th
     * entry represents the effective population size {@code g} generations
     * before the present.  The entries of the returned array are 0.0 for
     * {@code g < this.gMin()}.
     * @return the array of effective population sizes
     */
    public double[] ne() {
        return ne.clone();
    }

    /**
     * Returns the {@code BinnedIbd} object that was used to construct
     * {@code this}.
     * @return the {@code BinnedIbd} object that was used to construct
     * {@code this}
     */
    public BinnedIbd binnedIbd() {
        return binnedIbd;
    }

    /**
     * Returns the estimated sum of the lengths of observed IBD segments
     * whose most recent common ancestor is {@code g} generations before the
     * present generation.
     * @param g the number of generations before the present for the
     * most recent common ancestor of an IBD segment
     * @return the estimated sum of the lengths of observed IBD segments
     * whose most recent common ancestor is {@code g} generations before the
     * present generation
     * @throws IndexOutOfBoundsException if
     * {@code g < this.gMin() || g >= this.nGen()}
     */
    public double observedIbd(int g) {
        if (g<gMin) {
            throw new IndexOutOfBoundsException(String.valueOf(g));
        }
        return obs[g];
    }

    /**
     * Returns the product of the effective population size {@code g}
     * generations before the present and the expected sum of the lengths of
     * the IBD segments whose most recent common ancestor is {@code g}
     * generations before the present.
     * @param g the number of generations before the present for the
     * most recent common ancestor of an IBD segment
     * @return the product of the effective population size {@code g}
     * generations before the present and the expected sum of the lengths of
     * the IBD segments whose most recent common ancestor is {@code g}
     * generations before the present
     * @throws IndexOutOfBoundsException if
     * {@code g < this.gMin() || g >= this.nGen()}
     */
    public double expIbdTimeN(int g) {
        if (g<gMin) {
            throw new IndexOutOfBoundsException(String.valueOf(g));
        }
        return ne[g]*exp[g];
    }

    /**
     * Returns the chi-square goodness of fit statistic for the
     * observed and expected IBD.
     * @return the chi-square goodness of fit statistic
     */
    public double fit() {
        return fit;
    }

    private static double[][][] probs(double[] probTMRCA, int gMin,
            double tailN, BinnedIbd binnedIbd) {
        int nCensored = Ibd.CENSORED_LENGTH;
        int nIbdLengths = binnedIbd.nIbdLengths();
        double[][] gamma = gamma(probTMRCA, gMin, tailN, binnedIbd);
        double[][][] pi = new double[nCensored][probTMRCA.length][nIbdLengths];
        for (int len=0; len<nIbdLengths; ++len) {
            double length = binnedIbd.ibdLength(len);
            for (int g=gMin; g<probTMRCA.length; ++g) {
                double x = Math.exp(-length*g/50.0)*probTMRCA[g];
                double rate = g/50.0;
                pi[0][g][len] = x*rate*rate/gamma[0][len];
                pi[1][g][len] = x*rate/gamma[1][len];
                pi[2][g][len] = x/gamma[2][len];
            }
        }
        return pi;
    }

    private static double[][] gamma(double[] probTMRCA, int gMin, double tailN,
            BinnedIbd binnedIbd) {
        int nIbdLengths = binnedIbd.nIbdLengths();
        double[][] gamma = new double[Ibd.CENSORED_LENGTH][nIbdLengths];
        for (int g=gMin; g<probTMRCA.length; ++g) {
            double rate = g/50.0;
            for (int k=0; k<nIbdLengths; ++k) {
                double y2 = Math.exp(-binnedIbd.ibdLength(k)*rate)*probTMRCA[g];
                double y1 = rate*y2;
                double y0 = rate*y1;
                gamma[0][k] += y0;
                gamma[1][k] += y1;
                gamma[2][k] += y2;
            }
        }
        addGammaTailSum(probTMRCA, tailN, binnedIbd, gamma);
        return gamma;
    }

    private static void addGammaTailSum(double[] probTMRCA, double tailN,
            BinnedIbd binnedIbd, double[][] gamma) {
        int nIbdLengths = binnedIbd.nIbdLengths();
        double gMax = probTMRCA.length - 1;
        double beta = 0.5/tailN;
        double product = probTMRCA[probTMRCA.length-1] * (1 - 0.5/tailN)/beta;

        for (int j=0; j<nIbdLengths; ++j) {
            double alpha = binnedIbd.ibdLength(j)/50.0;
            double fraction = 1.0/(1.0 - (1.0 - beta)*Math.exp(-alpha));
            double x = product*beta*Math.exp(-alpha*(gMax+1))*fraction;
            gamma[0][j] += x*(gMax*gMax + fraction*(2*gMax - 1 + 2*fraction))/2500.0;
            gamma[1][j] += x*(gMax + fraction)/50.0;
            gamma[2][j] += x;
        }
    }

    private static double[] probTMRCA(double[] N, int minG) {
        double[] probTMRCA = new double[N.length];
        double product = 1.0f;
        for (int g=minG; g<N.length; ++g) {
            double beta = 0.5/N[g];
            probTMRCA[g] = product*beta;
            product *= (1.0 - beta);
        }
        return probTMRCA;
    }
}
