/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import blbutil.Const;
import blbutil.FileUtil;
import blbutil.Utilities;
import ints.IntList;
import java.io.File;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Locale;
import java.util.Random;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingDeque;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.TimeUnit;
import java.util.stream.IntStream;

/**
 * Class {@code NeMain} contains the entry point for the {@code IBDNe} program.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class NeMain {

    /**
     * The PROGRAM name and version.
     */
    public static final String PROGRAM = "ibdne.23Apr20.ae9.jar";

    /**
     * The COPYRIGHT string.
     */
    public static final String COPYRIGHT = "Copyright (C) 2015-2020 Brian L. Browning";

    private NeMain() {
        // private constructor to prevent instantiation
    }

    /**
     * Entry point to the {@code IBDNe} program. The command line
     * arguments are described in the string returned
     * by the {@code ibdne.NePar.usage()} method.
     *
     * @param args command line arguments
     */
    public static void main(String[] args) {
        Locale.setDefault(Locale.US);
        if (args.length==0) {
            Utilities.exit(NePar.usage());
        }
        long startNanoTime = System.nanoTime();
        NePar par = new NePar(args);
        try (PrintWriter log=FileUtil.printWriter(new File(par.out() + ".log"))) {
            long startNanos = System.nanoTime();
            Output.printStartInfo(par, log);
            Ibd ibd = new Ibd(new IbdReader(par));
            Output.printIbdInfo(ibd, startNanos, log);
            Ne[] results = multiThreadedAnalysis(par, ibd, par.seed(), log);

            Output.printFit(results, log);
            Output.printBoot(par, results);
            Output.printNe(par, results[0],
                    Arrays.copyOfRange(results, 1, results.length));
            Output.printEndInfo(startNanoTime, log);
        }
    }

    private static Ne estimateNe(NePar par, BinnedIbd binnedIbd, int[][] gStarts) {
        int gMin = par.gmin();
        double minNe = par.minne();
        double maxNe = par.maxne();
        HarmonicMean harmMean = new HarmonicMean(par.gmax()+1);
        double[] initNe = initNe(par);
	for (int[] gStarts1 : gStarts) {
            Ne ne = new Ne(initNe, gMin, binnedIbd);
            boolean finished = false;
            for (int j=0, nIts=par.nits(); j<nIts && finished==false; ++j) {
                Ne nextNe = updateNe(ne, gStarts1, minNe, maxNe);
                if((ne.fit() - nextNe.fit()) < 0.01){
                    finished = true;
                }
                ne = nextNe;
	    }
            harmMean.add(ne.ne());
	}
        return new Ne(harmMean.mean(), gMin, binnedIbd);
    }

    private static double[] initNe(NePar par) {
        return IntStream.rangeClosed(0, par.gmax())
            .mapToDouble(g -> par.initsize())
            .toArray();
    }

    private static Ne updateNe(Ne ne, int[] gStarts, double minNe,
            double maxNe) {
        int gMin = ne.gMin();
        double[] expIbd = new double[ne.nGen()];
        for (int g=gMin; g<ne.nGen(); ++g) {
            expIbd[g] = ne.expIbdTimeN(g);
        }
        int index = gStarts.length-1;
        int gStart = gStarts[index];
        int gEnd = ne.nGen();
        double[] neArray = ne.ne();
        setFinalInterval(ne, neArray, gStart, gEnd, minNe, maxNe);
        while (index > 0) {
            gEnd = gStart;
            gStart = gStarts[--index];
            setInterval(ne, neArray, expIbd, gStart, gEnd, minNe, maxNe);
        }
        assert gStart == gMin;
        return new Ne(neArray, ne.gMin(), ne.binnedIbd());
    }

    private static void setFinalInterval(Ne ne, double[] neArray, int gStart,
            int gEnd, double minNe, double maxNe) {
        double estNe = estPopSize(ne, gStart, gEnd);
        if (estNe < minNe) {
            estNe = minNe;
        }
        if (estNe > maxNe) {
            estNe = maxNe;
        }
        Arrays.fill(neArray, gStart, gEnd, estNe);
    }

    private static double estPopSize(Ne ne, int gStart, int gEnd) {
        double numSum = 0;
        double denSum = 0;
        for (int g=gStart; g<gEnd; ++g) {
            numSum += ne.expIbdTimeN(g);
            denSum += ne.observedIbd(g);
        }
        return numSum/denSum;
    }

    private static void setInterval(Ne ne, double[] neArray, double[] expIbd,
            int gStart, int gEnd, double minNe, double maxNe) {
        if ((gStart+1)==gEnd) {
            neArray[gStart] = ne.expIbdTimeN(gStart)/ne.observedIbd(gStart);
            enforceBounds(neArray, gStart, minNe, maxNe);
        }
        else {
            double c = sumObservedIbd(ne, gStart, gEnd) * neArray[gEnd];
            double r = solve(expIbd, gStart, gEnd, c);
            if (Double.isNaN(r) || Math.abs(r)>2) {
                for (int g=gStart; g<gEnd; ++g) {
                    neArray[g] = ne.expIbdTimeN(g)/ne.observedIbd(g);
                    enforceBounds(neArray, g, minNe, maxNe);
                }
            }
            else {
                for (int g=gStart; g<gEnd; ++g) {
                    neArray[g] = neArray[gEnd]*Math.exp(r*(gEnd-g));
                    enforceBounds(neArray, g, minNe, maxNe);
                }
            }
        }
        if (gStart==ne.gMin()) {
            double r = Math.log(neArray[gStart]/neArray[gStart+1]);
            for (int g=0; g<gStart; ++g) {
                neArray[g] = neArray[gStart]*Math.exp(r*(gStart-g));
                enforceBounds(neArray, g, minNe, maxNe);
            }
        }
    }

    private static void enforceBounds(double[] ne, int g, double minNe,
            double maxNe) {
        if (ne[g] < minNe) {
            ne[g] = minNe;
        }
        if (ne[g] > maxNe) {
             ne[g] = maxNe;
        }
    }

    private static double sumObservedIbd(Ne ne, int gStart, int gEnd) {
        double sum = 0.0;
        for (int g=gStart; g<gEnd; ++g) {
            sum += ne.observedIbd(g);
        }
        return sum;
    }

    private static double solve(double[] expIbd, int gStart, int gEnd, double c) {
        int minIts = 10;
        int maxIts = 100;
        double changeThreshold = 0.001;
        double change = Double.MAX_VALUE;
        double r = 0d;
        int it = 0;
        while (it < minIts || (it < maxIts && change > changeThreshold)) {
            double lastR = r;
            r = nextR(lastR, c, expIbd, gStart, gEnd);
            change = Math.abs(r - lastR);
            ++it;
        }
        return (change > changeThreshold) ? Double.NaN : r;
    }

    private static double nextR(double r, double c, double[] expIbd, int gStart,
            int gEnd) {
        double f = c;
        double fp = 0.0;
        for (int g=gStart; g<gEnd; ++g) {
            int k = g - gEnd;
            double x = expIbd[g]*Math.exp(r*k);
            f -= x;
            fp -= k*x;
        }
        return r - f/fp;
    }

    private static int[][] gStarts(NePar par) {
        int gMin = par.gmin();
        int nGen = par.gmax() + 1;
        int minOffset = par.offset() + 1;
        int maxOffset = par.offset() + par.group();
        int laterGroupSize = par.group();

        int[][] gStarts = new int[maxOffset - minOffset + 1][];
        IntList starts = new IntList(3 + nGen/laterGroupSize);
        for (int size=minOffset; size<=maxOffset; ++size) {
            starts.clear();
            starts.add(gMin);
            int gStart = gMin + size;
            while (gStart < nGen) {
                starts.add(gStart);
                gStart += laterGroupSize;
            }
            gStarts[size-minOffset] = starts.toArray();
        }
        return gStarts;
    }

    private static Ne[] multiThreadedAnalysis(NePar par, Ibd ibd,
            long seed, PrintWriter log) {
        int capacity = par.nboots()+1;
        BlockingQueue<Task> qIn = new ArrayBlockingQueue<>(capacity);
        BlockingDeque<Ne> qOut = new LinkedBlockingDeque<>(capacity);
        Random random  = new Random(seed);

        ExecutorService es = Executors.newFixedThreadPool(par.nthreads());
        try {
            qIn.put(new Task(ibd, random.nextLong(), false));
            for (int j=0; j<par.nboots(); ++j) {
                qIn.put(new Task(ibd, random.nextLong(), true));
            }
            for (int j=0; j<par.nthreads(); ++j) {
                es.submit(new EstimateNe(par, qIn, qOut, log));
            }
            es.shutdown();
            es.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
        }
        catch (Throwable t) {
            Utilities.exit(t);
        }
        return qOut.toArray(new Ne[0]);
    }

    private static class EstimateNe implements Runnable {

        private final NePar par;
        private final BlockingQueue<Task> qIn;
        private final BlockingDeque<Ne> qOut;
        private final PrintWriter log;

        public EstimateNe(NePar par, BlockingQueue<Task> qIn,
                BlockingDeque<Ne> qOut, PrintWriter log) {
            if (par==null) {
                throw new NullPointerException("ne==null");
            }
            if (qIn==null) {
                throw new NullPointerException("qIn==null");
            }
            if (qOut==null) {
                throw new NullPointerException("qOut==null");
            }
            if (log==null) {
                throw new NullPointerException("log==null");
            }
            this.par = par;
            this.qIn = qIn;
            this.qOut = qOut;
            this.log = log;
        }

        @Override
        public void run() {
            try {
                int[][] gStarts = gStarts(par);
                Task task = qIn.poll();
                while (task!=null) {
                    Random random = new Random(task.seed);
                    Ibd ibd = task.ibd;
                    if (task.bootstrap) {
                        ibd = ibd.bootstrapSample(random.nextLong());
                        BinnedIbd binnedIbd = new BinnedIbd(ibd);
                        Ne ne = estimateNe(par, binnedIbd, gStarts);
                        qOut.putLast(ne);
                    }
                    else {
                        long start = System.nanoTime();
                        BinnedIbd binnedIbd = new BinnedIbd(ibd);
                        Ne ne = estimateNe(par, binnedIbd, gStarts);
                        qOut.putFirst(ne);
                        long elapsedNanos = System.nanoTime() - start;

                        Utilities.duoPrint(log, Const.nl
                                + "Estimating Ne [original data]      ");
                        Utilities.duoPrintln(log, "Done (elapsed time: "
                                + Utilities.elapsedNanos(elapsedNanos) + ")");
                        Utilities.duoPrintln(log,
                                "Estimating Ne [bootstrap samples]");
                    }
                    task = qIn.poll();
                }
            } catch (Throwable t) {
                Utilities.exit(t);
            }
        }
    }

    private static class Task {
        private final Ibd ibd;
        private final long seed;
        private final boolean bootstrap;

        public Task(Ibd ibd, long seed, boolean bootstrapSample) {
            if (ibd==null) {
                throw new NullPointerException("ibd==null");
            }
            this.ibd = ibd;
            this.seed = seed;
            this.bootstrap = bootstrapSample;
        }
    }
}
