/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import blbutil.Const;
import blbutil.Validate;
import java.io.File;
import java.text.DecimalFormat;
import java.util.Map;

/**
 * <p>Class {@code NePar} represents the analysis parameters for the
 * {@code IBDNe} program.
 * </p>
 * Instances of class {@code NePar} are immutable.
 *
 * @author Brian L Browning {@code <browning@uw.edu>}
 */
public final class NePar {

    // hard-coded parameters
    private final double minne = 10;        // min effective population size; must be > 0
    private final double maxne = 1.0e100;   // max effective population size; must be > minne
    private final double binwidth = 0.05;   // width of IBD bin range in cM; must be >0
    private final double initsize = 100;    // constant population size used to initialize algorithm; must be >0
    private final double conf = 0.95;       // confidence level; must be >0 and <1
    private final double maxsharing = 0.4;  // max proportion of genome shared; must be >0 and <1

    private final double trimprop = 0.03f;  // trim proportion for trimmed mean and variance; must be >=0 and <0.5
    private final double interval = 0.25f;  // cM interval between positions with IBD segment counts; must be >0
    private final double maxdev = 10f;      // max std dev in IBD segments at a locus; must be >0
    private final int offset = 4;           // controls size of first group of generations
    private final int group = 8;            // number of generations to group when estimaitng growth rate.

    // documented parameters
    private final File map;
    private final String out;
    private final int nthreads;
    private final boolean filtersamples;
    private final long npairs;
    private final int nits;
    private final int nboots;
    private final double mincm;
    private final double trimcm;
    private final int gmin;
    private final int gmax;
    private final long seed;

    // undocumented parameters
    private final double minregion;

    // default documented parameters
    private static final int     DEF_NTHREADS = Integer.MAX_VALUE;
    private static final boolean DEF_FILTERSAMPLES = true;
    private static final int     DEF_NPAIRS = 0;
    private static final int     DEF_NITS = 1000;
    private static final int     DEF_NBOOTS = 80;
    private static final double  DEF_MINCM = 2f;
    private static final double  DEF_TRIMCM = 0.2f;
    private static final int     DEF_GMIN = 2;
    private static final int     DEF_GMAX = Integer.MAX_VALUE;
    private static final int     DEF_SEED = -99999;

    // default undocumented parameters
    private static final double  DEF_MINREGION = 50f;

    /**
     * Constructs an {@code NePar} instance from the command line
     * arguments. Each element in the specified array must have the
     * format "[name]=[value]".  The string returned by the {@code usage()}
     * method contains a description of the command line arguments.
     *
     * @param args the command line arguments.
     * @throws IllegalArgumentException if an element of the specified
     * array has an unrecognized name or an invalid value.
    */
    public NePar(String[] args) {
        int IMAX = Integer.MAX_VALUE;
        long LMIN = Long.MIN_VALUE;
        long LMAX = Long.MAX_VALUE;
        double FMIN = Float.MIN_VALUE;
        double FMAX = Float.MAX_VALUE;

        Map<String, String> argsMap = Validate.argsToMap(args, '=');

        // documented parameters
        map = Validate.getFile(Validate.stringArg("map", argsMap, true, null,
                null));
        out = Validate.stringArg("out", argsMap, true, null, null);
        mincm = Validate.doubleArg("mincm", argsMap, false, DEF_MINCM, FMIN, FMAX);
        minregion = Validate.doubleArg("minregion", argsMap, false, DEF_MINREGION, FMIN, FMAX);
        trimcm = Validate.doubleArg("trimcm", argsMap, false, DEF_TRIMCM, 0.0f, FMAX);
        filtersamples = Validate.booleanArg("filtersamples", argsMap, false, DEF_FILTERSAMPLES);
        npairs = Validate.longArg("npairs", argsMap, false, DEF_NPAIRS, 0, LMAX);
        nits = Validate.intArg("nits", argsMap, false, DEF_NITS, 1, IMAX);
        nboots = Validate.intArg("nboots", argsMap, false, DEF_NBOOTS, 0, IMAX);
        seed = Validate.longArg("seed", argsMap, false, DEF_SEED, LMIN, LMAX);
        nthreads = modNthreads(Validate.intArg("nthreads", argsMap, false, DEF_NTHREADS, 1, IMAX));

        gmin = Validate.intArg("gmin", argsMap, false, DEF_GMIN, 1, 2);
        gmax = modGMax(Validate.intArg("gmax", argsMap, false, DEF_GMAX, 4, IMAX));

        if (filtersamples && npairs!=DEF_NPAIRS) {
            String s = "ERROR: \"filtersamples=false\" is required if \"npairs\" is > 0";
            throw new IllegalArgumentException(s);
        }

        Validate.confirmEmptyMap(argsMap);
    }

    private static int modNthreads(int nthreads) {
        if (nthreads==DEF_NTHREADS) {
            return Runtime.getRuntime().availableProcessors();
        }
        else {
            return nthreads;
        }
    }

    private int modGMax(int gmax) {
        return (gmax!=DEF_GMAX) ? gmax : (int) Math.ceil(600.0/mincm);
    }

    /**
     * Returns a string describing the command line arguments.
     * The format of the returned string is unspecified and subject to change.
     * @return a string describing the command line arguments.
     */
    public static String usage() {
        String nl = Const.nl;
        DecimalFormat df3 = new DecimalFormat("0.0##");
        return  "Program: " + NeMain.PROGRAM + nl
                + NeMain.COPYRIGHT + nl
                + nl
                + "usage: cat [ibd] | java -jar " + NeMain.PROGRAM + " [args]" + nl
                + nl
                + "where: " + nl
                + "  [ibd]  = a space-separated list of hap-IBD output IBD files" + nl
                + "  [args] = a space-separated list of algorithm parameters" + nl
                + nl
                + "Algorithm Parameters: " + nl
                + "  map=<PLINK-format genetic map with cM distances>  (required)" + nl
                + "  out=<output file prefix>                          (required)" + nl
                + "  nthreads=<number of computational threads>        (default=[all CPU cores])" + nl
                + "  filtersamples=<true/false>                        (default=" + DEF_FILTERSAMPLES + ")" + nl
                + "  npairs=<number of unordered haplotype pairs>      (default=" + DEF_NPAIRS + ")" + nl
                + "  nits=<number of iterations>                       (default=" + DEF_NITS + ")" + nl
                + "  nboots<number of bootstrap samples>               (default=" + DEF_NBOOTS + ")" + nl
                + "  mincm=<min cM length of an IBD segment>           (default=" + df3.format(DEF_MINCM) + ")" + nl
                + "  trimcm=<cM to trim from ends of each region>      (default=" + df3.format(DEF_TRIMCM) +")"  + nl
                + "  gmin=<min generations before present (1 or 2)>    (default=" + DEF_GMIN +")" + nl
                + "  gmax=<max number of generations before present>   (default=[600/mincm])" +nl
                + "  seed=<seed for random number generator>           (default=" + DEF_SEED + ")" + nl;
    }

    // documented parameters

    /**
     * Returns the map parameter. The map parameter is a PLINK-format
     * genetic map file with genetic distances in centiMorgan units.
     * @return the map parameter.
     */
    public File map() {
        return map;
    }

    /**
     * Returns the out parameter. The out parameter is the prefix for
     * output filenames.
     * @return the out parameter.
     */
    public String out() {
        return out;
    }

    /**
     * Returns the nthreads parameter. The nthreads parameter is the number
     * of computational threads that will be used for the analysis.
     * @return the nthreads parameter
     */
    public int nthreads() {
        return nthreads;
    }

    /**
     * Returns the filtersamples parameter.  The filtersamples parameter
     * controls whether closely-related samples are excluded
     * ({@code filtersamples=true}) or included ({@code filtersamples=false})
     * @return the filtersamples parameter.
     */
    public boolean filtersamples() {
        return filtersamples;
    }

    /**
     * Returns the npairs parameter.  The npairs parameter is the number
     * of haplotype pairs that produce the observed IBD.  If
     * {@code this.npairs()==0} then the number of haplotype pairs will
     * be determined from the input data to be {@code N*(N-2)/2} where
     * {@code N} is two times the number of distinct sample identifiers
     * in the input IBD file.
     *
     * @return the npairs parameter
     */
    public long npairs() {
        return npairs;
    }

    /**
     * Returns the nits parameter. The nits parameter is the number of
     * algorithm iterations to use when estimating effective population size
     * at each generation before the present.
     * @return the nits parameter.
     */
    public int nits() {
        return nits;
    }

    /**
     * Returns the nboots parameter. The nboots parameter is the number
     * of bootstrap samples to use when estimating confidence intervals
     * for the effective population size at each generation for the
     * present.
     * @return the nboots parameter.
     */
    public int nboots() {
        return nboots;
    }

    /**
     * Returns the mincm parameter. The mincm parameter is the minimum
     * length threshold in centiMorgan (cM) units of the IBD segments
     * that are analyzed.
     * @return the mincm parameter.
     */
    public double mincm() {
        return mincm;
    }

    /**
     * Returns the trimcm parameter. The trimcm parameter is the length in
     * centiMorgan (cM) units that is trimmed from each end of each chromosome.
     * @return the trimcm parameter.
     */
    public double trimcm() {
        return trimcm;
    }

    /**
     * Returns the gmin parameter.  The gin parameter is the minimum
     * generation before the present that will have its effective population
     * size estimated.
     * @return the gmin parameter
     */
    public int gmin() {
        return gmin;
    }

    /**
     * Returns the gmax parameter.  The gmax parameter is the maximimum
     * generation before the present that will have its effective population
     * size estimated.
     * @return the gmax parameter
     */
    public int gmax() {
        return gmax;
    }

    /**
     * Returns the seed parameter. The seed parameter is the seed for
     * the random number generator.
     * @return the seed parameter
     */
    public long seed() {
        return seed;
    }

    // undocumented parameters

    /**
     * Returns the minregion parameter. The minregion parameter is the minimum
     * length of a continuous genomic region with IBD segment data.
     *
     * @return the minregion parameter
     */
    public double minregion() {
        return minregion;
    }

    // hard-coded parameters

    /**
     * Returns the minne parameter. The minne parameter is a lower bound
     * on the estimated effective population size in each generation.
     * @return the minne parameter
     */
    public double minne() {
        return minne;
    }

    /**
     * Returns the maxne parameter. The maxne parameter is an uppper bound
     * on the estimated effective population size in each generation.
     * @return the maxne parameter
     */
    public double maxne() {
        return maxne;
    }

    /**
     * Returns the binwidth parameter. The binwidth parameter is the
     * difference between the maximum and minimum lengths of the IBD segments
     * that are assigned to a bin.
     * @return the binwidth parameter
     */
    public double binwidth() {
        return binwidth;
    }

    /**
     * Returns the initsize parameter. The initsize parameter is constant
     * effective population size used to initialize the iterative algorithm.
     * @return the initsize parameter
     */
    public double initsize() {
        return initsize;
    }

    /**
     * Returns the conf parameter. The conf parameter is the confidence level
     * used for estimating upper and lower bounds on the effective population
     * size at each generation.
     * @return the conf parameter
     */
    public double conf() {
        return conf;
    }

    /**
     * Returns the maxsharing parameter. If {@code this.filtersamples()==true}
     * and the sum of IBD segments for a pair of samples exceeds this
     * proportion of the genome, all IBD segments for the pair of samples
     * will be excluded.  The maxsharing parameter has no effect if
     * {@code this.filtersamples()==false}.
     * @return the maxsharing parameter
     */
    public double maxsharing() {
        return maxsharing;
    }

    /**
     * Returns the trimprop parameter. The trimprop parameter is the
     * trim proportion when computing a trimmed mean and variance
     *
     * @return the trimprop parameter
     */
    public double trimprop() {
        return trimprop;
    }

    /**
     * Returns the interval parameter. The interval parameter is the
     * centiMorgan length between positions at which IBD segments are counted.
     *
     * @return the interval parameter.
     */
    public double interval() {
        return interval;
    }

    /**
     * Returns the maxdev parameter. The maxdev parameter is the maximum
     * permitted standard deviations in the number of IBD segments at a locus.
     * Genomic loci with an extreme number of IBD segments will be excluded.
     *
     * @return the maxdev parameter
     */
    public double maxdev() {
        return maxdev;
    }

    /**
     * Returns the offset parameter.  The offset parameter controls the
     * number of generations that are in the first group of generations
     * when estimating the effective population size with an exponential
     * growth/decay model.  The number of generations in the first group
     * is equal to the sum of the offset parameter and a random variable
     * that is uniformly distributed over the integers 1, 2, ...,
     * {@code this.group()}.
     * @return the group parameter
     */
    public int offset() {
        return offset;
    }

    /**
     * Returns the group parameter.  The group parameter is the number of
     * consecutive generations to group when estimating the effective
     * population size with an exponential growth/decay model.
     * @return the group parameter
     */
    public int group() {
        return group;
    }
}
