/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdne;

import beagleutil.SampleIds;
import blbutil.Const;
import blbutil.FileUtil;
import blbutil.Utilities;
import java.io.File;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.Arrays;

/**
 * Class {@code Output} contains static methods for printing
 * log messages from the {@code IBDNe} PROGRAM.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
class Output {

    private static final DecimalFormat df0 = new DecimalFormat("0");
    private static final DecimalFormat df1 = new DecimalFormat("0.0");
    private static final DecimalFormat df3 = new DecimalFormat("0.0##");

    private Output() {
        // private constructor to prevent instantiation
    }

    static void printStartInfo(NePar par, PrintWriter log) {
        StringBuilder sb = new StringBuilder(300);
        long maxMemory = Runtime.getRuntime().maxMemory();
        sb.append("Program           :  ");
        sb.append(NeMain.PROGRAM);
        sb.append(Const.nl);
        sb.append("Author            :  ");
        sb.append(NeMain.COPYRIGHT);
        sb.append(Const.nl);
        sb.append("Available memory  :  ");
        if (maxMemory != Long.MAX_VALUE) {
            long maxMb = maxMemory / (1024 * 1024);
            sb.append(maxMb);
            sb.append(" MB");
        } else {
            sb.append("[no limit])");
        }
        sb.append(Const.nl);
        sb.append("Start Time        :  ");
        sb.append(Utilities.timeStamp());
        sb.append(Const.nl);
        sb.append(Const.nl);
        sb.append(currentParameters(par));
        sb.append(Const.nl);
        sb.append("Reading IBD segments               ");
        Utilities.duoPrint(log, sb);
    }

    private static String currentParameters(NePar par) {
        StringBuilder sb = new StringBuilder(150);
        sb.append("Parameters");
        sb.append(Const.nl);
        sb.append("  map             :  ");
        sb.append(par.map());
        sb.append(Const.nl);
        sb.append("  out             :  ");
        sb.append(par.out());
        sb.append(Const.nl);
        sb.append("  mincm           :  ");
        sb.append(par.mincm());
        sb.append(Const.nl);
        sb.append("  minregion       :  ");
        sb.append(par.minregion());
        sb.append(Const.nl);
        sb.append("  trimcm          :  ");
        sb.append(df3.format(par.trimcm()));
        sb.append(Const.nl);
        sb.append("  filtersample    :  ");
        sb.append(par.filtersamples());
        sb.append(Const.nl);
        sb.append("  npairs          :  ");
        sb.append(par.npairs() == 0 ? "data dependent" : par.npairs());
        sb.append(Const.nl);
        sb.append("  nits            :  ");
        sb.append(par.nits());
        sb.append(Const.nl);
        sb.append("  nboots          :  ");
        sb.append(par.nboots());
        sb.append(Const.nl);
        sb.append("  gmin            :  ");
        sb.append(par.gmin());
        sb.append(Const.nl);
        sb.append("  gmax            :  ");
        sb.append(par.gmax());
        sb.append(Const.nl);
        sb.append("  seed            :  ");
        sb.append(par.seed());
        sb.append(Const.nl);
        sb.append("  nthreads        :  ");
        sb.append(par.nthreads());
        sb.append(Const.nl);
        return sb.toString();
    }

    static void printIbdInfo(Ibd ibd, long startNanos, PrintWriter log) {
        long elapsedNanos = System.nanoTime() - startNanos;
        StringBuilder sb = new StringBuilder(300);
        sb.append("Done (elapsed time: ");
        sb.append(Utilities.elapsedNanos(elapsedNanos));
        sb.append(")");
        sb.append(Const.nl);
        sb.append("Samples           : ");
        sb.append(SampleIds.instance().size());
        sb.append(Const.nl);
        sb.append("segments read     : ");
        sb.append(ibd.nInputSegments());
        sb.append(Const.nl);
        sb.append("segments analyzed : ");
        sb.append(ibd.nSegmentsAnalyzed());
        Utilities.duoPrintln(log, sb);
    }

    static void printEndInfo(long startNanoTime, PrintWriter log) {
        long elapsedNanoTime = System.nanoTime() - startNanoTime;
        StringBuilder sb = new StringBuilder(300);
        sb.append(Const.nl);
        sb.append("Elapsed Time      :  ");
        sb.append(Utilities.elapsedNanos(elapsedNanoTime));
        sb.append(Const.nl);
        sb.append("End Time          :  ");
        sb.append(Utilities.timeStamp());
        Utilities.duoPrintln(log, sb);
    }

    static void printBoot(NePar par, Ne[] results) {
        File outFile = new File(par.out() + ".boot");
        try (final PrintWriter out = FileUtil.printWriter(outFile)) {
            for (Ne ne : results) {
                for (int g = 0, n = ne.nGen(); g < n; ++g) {
                    out.print(format(df0, ne.ne(g)));
                    out.print(g < (n - 1) ? Const.tab : Const.nl);
                }
            }
        }
    }

    static void printNe(NePar par, Ne ne, Ne[] bsProbG) {
        boolean printConfInterval = bsProbG.length > 0;
        double confLevel = par.conf();
        double minProp = (1.0F - confLevel) / 2.0F;
        double maxProp = 1.0F - minProp;
        DecimalFormat dfE2 = new DecimalFormat("0.00E0");
        assert confLevel > 0.0F && confLevel < 1.0F;
        try (final PrintWriter out = FileUtil.printWriter(new File(par.out() + ".ne"))) {
            printNeHeader(printConfInterval, confLevel, out);
            for (int g = 0, n = ne.nGen(); g < n; ++g) {
                out.print(g);
                out.print(Const.tab);
                out.print(format(dfE2, ne.ne(g)));
                if (printConfInterval) {
                    double[] sorted = getSortedArray(bsProbG, g);
                    double lwr = getPercentile(minProp, sorted);
                    double upr = getPercentile(maxProp, sorted);
                    out.print(Const.tab);
                    out.print(format(dfE2, lwr));
                    out.print(Const.tab);
                    out.print(format(dfE2, upr));
                }
                out.println();
            }
        }
    }

    private static void printNeHeader(boolean printConfIntervals,
            double confLevel, PrintWriter out) {
        out.print("GEN");
        out.print(Const.tab);
        out.print("NE");
        if (printConfIntervals) {
            out.print(Const.tab);
            out.print("LWR-" + df0.format(100 * confLevel) + "%CI");
            out.print(Const.tab);
            out.print("UPR-" + df0.format(100 * confLevel) + "%CI");
        }
        out.println();
    }

    static void printFit(Ne[] results, PrintWriter log) {
        Utilities.duoPrint(log, Const.nl);
        Utilities.duoPrint(log, "Original  fit ");
        Utilities.duoPrintln(log, df1.format(results[0].fit()));
        for (int j = 1; j < results.length; ++j) {
            Utilities.duoPrint(log, "Bootstrap fit ");
            Utilities.duoPrintln(log, df1.format(results[j].fit()));
        }
    }

   private static String format(DecimalFormat decFormat, double d) {
        if (Double.isFinite(d)==false) {
            return String.valueOf(d);
        }
        else {
            return decFormat.format(d);
        }
    }

    private static double[] getSortedArray(Ne[] bsNe, int g) {
        double[] rankSorted = new double[bsNe.length];
        for (int j=0; j<rankSorted.length; ++j) {
            rankSorted[j] = bsNe[j].ne(g);
        }
        Arrays.sort(rankSorted);
        return rankSorted;
    }

    /* implements "NIST" percentile method" */
    private static double getPercentile(double prop, double[] sorted) {
        if (sorted.length==0) {
            throw new IllegalArgumentException("sorted.length==0");
        }
        if (prop<0.0 || prop>1.0) {
            throw new IllegalArgumentException("prop=" + prop);
        }
        int lastIndex = sorted.length-1;
        double n = prop*(sorted.length+1) - 1;
        int index = (int) Math.floor(n);
        double remainder = n - Math.floor(n);
        if (index<0) {
            return sorted[0];
        }
        else if (index>=lastIndex) {
            return sorted[lastIndex];
        }
        else {
            return sorted[index] + remainder*(sorted[index+1] - sorted[index]);
        }
    }
}
