/*
 * Copyright 2015-2016 Brian L. Browning
 *
 * This file is part of IBDNe
 *
 * IBDNE is licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ibdutil;

import beagleutil.ChromIds;
import beagleutil.IntInterval;
import beagleutil.SampleIds;
import blbutil.Const;
import blbutil.StringUtil;
import java.text.DecimalFormat;

/**
 * Class {@code StringIdSegment} represents an immutable IBD segment with
 * string sample identifiers.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public final class StringIdSegment implements IntInterval,
        Comparable<StringIdSegment> {

    private final int sampleIndex1;
    private final int hap1;
    private final int sampleIndex2;
    private final int hap2;
    private final int chromIndex;
    private final int start;
    private final int end;
    private final float score;
    private final float length;

    /**
     * Constructs a {@code StringIdSegment} instance.  Segments which
     * do not have a specified length should have length set to
     * {@code Float.NaN}.
     *
     * @param sample1 the first sample identifier
     * @param hap1 the one-based index of the IBD haplotype in the first
     * sample (0 = unknown)
     * @param sample2 the second sample identifier
     * @param hap2 the one-based index of the IBD haplotype in the second
     * sample (0 = unknown)
     * @param chrom the chromosome identifier
     * @param start the starting position (inclusive)
     * @param end the ending position (inclusive)
     * @param score the score for the IBD segment
     * @param length the length of the IBD segment
     *
     * @throws NullPointerException if any string is {@code null}
     * @throws IllegalArgumentException if {@code start > end}
     * @throws IllegalArgumentException if
     * {@code hap1 < 0 || hap1 > 2 || hap2 < 0 || hap2 > 2}
     */
    public StringIdSegment(String sample1, int hap1, String sample2, int hap2,
            String chrom, int start, int end, float score, float length) {
        if (start > end) {
            String s = "start=" + start + " end=" + end;
            throw new IllegalArgumentException(s);
        }
        checkHaplotype(hap1);
        checkHaplotype(hap2);
        this.sampleIndex1 = SampleIds.instance().getIndex(sample1);
        this.hap1 = hap1;
        this.sampleIndex2 = SampleIds.instance().getIndex(sample2);
        this.hap2 = hap2;
        this.chromIndex = ChromIds.instance().getIndex(chrom);
        this.start = start;
        this.end = end;
        this.score = score;
        this.length = length;
    }

    private static void checkHaplotype(int hap) {
        if (hap!=0 && hap!=1 && hap!=2) {
            throw new IllegalArgumentException("invalid hap: " + hap);
        }
    }

    /**
     * Returns a new {@code StringIdSegment} object initialized to the
     * to the value represented by the specified {@code String}.
     * The specified string must contain exactly 8 or 9 white-space delimited
     * fields representing 1) the first sample identifier, 2) the index of the
     * IBD haplotype for the first sample (0, 1 or 2), 3) the second sample
     * identifier, 4) the index of the IBD haplotype for the second sample
     * (0, 1 or 2), 5) the chromosome, 6) the starting marker position (inclusive),
     * 7) the ending marker position (inclusive), and 8) the IBD segment score,
     * and 9) the IBD segment length (optional).
     *
     * @param line the string to be parsed
     *
     * @return a new {@code StringIdSegment} object initialized to the
     * value represented by the specified {@code String}
     *
     * @throws NullPointerException if {@code line==null}
     * @throws IllegalArgumentException if the specified line does
     * not contain exactly 8 or 9 white-space delimited fields
     * @throws IllegalArgumentException if the starting marker position
     * (field 6) is greater than the ending marker position (field 7)
     * @throws IllegalArgumentException if the haplotype indices in
     * fields 2 and 4 are not 0, 1, or 2
     * @throws NumberFormatException if the second, fourth, sixth, and
     * seventh white-space delimited fields are not parsable integers, or
     * if the eighth or (optional) ninth white-space delimited fields are not
     * parseable floats
     */
    public static StringIdSegment parse(String line) {
        String[] fields = StringUtil.getFields(line);
        if (fields.length < 8 ||  fields.length > 9) {
            String s = "Line describing IBD segment does not contain 8 or 9 "
                    + "white-space delimited fields: " + line;
            throw new IllegalArgumentException(s);
        }
        String id1 = fields[0];
        int hap1 = Integer.parseInt(fields[1]);
        String id2 = fields[2];
        int hap2 = Integer.parseInt(fields[3]);
        String chrom = fields[4];
        int start = Integer.parseInt(fields[5]);
        int end = Integer.parseInt(fields[6]);
        float score = Float.parseFloat(fields[7]);
        float length = fields.length==8 ? Float.NaN : Float.parseFloat(fields[8]);
        return new StringIdSegment(id1, hap1, id2, hap2, chrom, start,
                end, score, length);
    }

    /**
     * Returns the first sample identifier.
     * @return the first sample identifier.
     */
    public String sample1() {
        return SampleIds.instance().id(sampleIndex1);
    }

    /**
     * Returns the first sample identifier index.
     * @return the first sample identifier index.
     */
    public int sampleIndex1() {
        return sampleIndex1;
    }

    /**
     * Returns the one-based index of the IBD haplotype in the first sample
     * (0 = unknown).
     * @return the one-based index of the IBD haplotype in the first sample.
     */
    public int hap1() {
        return hap1;
    }

    /**
     * Returns the second sample identifier.
     * @return the second sample identifier.
     */
    public String sample2() {
        return SampleIds.instance().id(sampleIndex2);
    }

    /**
     * Returns the second sample identifier index.
     * @return the second sample identifier index.
     */
    public int sampleIndex2() {
        return sampleIndex2;
    }

    /**
     * Returns the one-based index of the IBD haplotype in the second sample
     * (0 = unknown).
     * @return the one-based index of the IBD haplotype in the second sample.
     */
    public int hap2() {
        return hap2;
    }

    /**
     * Returns the chromosome.
     * @return the chromosome.
     */
    public String chrom() {
        return ChromIds.instance().id(chromIndex);
    }

    /**
     * Returns the chromosome index
     * @return the chromosome index.
     */
    public int chromIndex() {
        return chromIndex;
    }

    /**
     * Returns the starting marker position (inclusive).
     * @return the starting marker position (inclusive).
     */
    @Override
    public int start() {
        return start;
    }

    /**
     * Returns the ending marker position (inclusive).
     * @return the ending marker position (inclusive).
     */
    @Override
    public int inclEnd() {
        return end;
    }

     /**
     * Returns the IBD segment score.
     * @return the IBD segment score.
     */
    public float score() {
        return score;
    }

    /**
     * Returns the IBD segment length.
     * @return the IBD segment length
     */
    public float length() {
        return length;
    }

    /**
     * Returns a tab-delimited string whose first eight ordered fields are:
     * {@code this.sample1()}, {@code this.hap1()}, {@code this.sample2()},
     * {@code this.hap2()}, {@code this.chrom()}, {@code this.start()},
     * {@code this.end()}, and {@code this.score()}.  If
     * {@code (Float.isNaN(length) == false)}, the returned tab-delimited
     * string includes {@code this.length()} as the ninth field.
     * @return a string representation of {@code this}.
     */
    @Override
    public String toString() {
        char tab = Const.tab;
        StringBuilder sb = new StringBuilder();
        sb.append(this.sample1());
        sb.append(tab);
        sb.append(hap1);
        sb.append(tab);
        sb.append(this.sample2());
        sb.append(tab);
        sb.append(hap2);
        sb.append(tab);
        sb.append(this.chrom());
        sb.append(tab);
        sb.append(start);
        sb.append(tab);
        sb.append(end);
        sb.append(tab);
        sb.append(score);
        if (Float.isNaN(length)==false) {
            sb.append(tab);
            sb.append(length);
        }
        return sb.toString();
    }

    /**
     * Returns a tab-delimited string whose ordered fields are:
     * {@code this.sample1(), this.hap1(), this.sample2(), this.hap2(),
     * this.chrom(), this.start(), this.end(), and this.score()}.
     * @param df a {@code DecimalFormat} object to use for formatting
     * the IBD segment score.
     * @return a string representation of {@code this}.
     */
    public String toString(DecimalFormat df) {
        char tab = Const.tab;
        StringBuilder sb = new StringBuilder();
        sb.append(this.sample1());
        sb.append(tab);
        sb.append(hap1);
        sb.append(tab);
        sb.append(this.sample2());
        sb.append(tab);
        sb.append(hap2);
        sb.append(tab);
        sb.append(this.chrom());
        sb.append(tab);
        sb.append(start);
        sb.append(tab);
        sb.append(end);
        sb.append(tab);
        sb.append(df.format(score));
        return sb.toString();
    }

    /**
     * Returns {@code true} if this object and the specified object
     * are both instances of {@code StringIdSegment}, and have
     * identical first and second sample identifiers and IBD haplotype
     * indices, identical chromosomes, identical starting and ending marker
     * indices, and identical IBD scores (compared using
     * {@code Float.compare()}), and returns {@code false} otherwise.
     * @param o the reference object with which to compare.
     * @return {@code true} if this object is the equal to the specified
     * object and returns {@code false} otherwise.
     */
    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if ((o instanceof StringIdSegment) == false) {
            return false;
        }
        StringIdSegment other = (StringIdSegment) o;
        if (this.sampleIndex1 != other.sampleIndex1) {
            return false;
        }
        if (this.hap1!=other.hap1) {
            return false;
        }
        if (this.sampleIndex2 != other.sampleIndex2) {
            return false;
        }
        if (this.hap2!=other.hap2) {
            return false;
        }
        if (this.chromIndex != other.chromIndex) {
            return false;
        }
        if (this.start != other.start) {
            return false;
        }
        if (this.end != other.end) {
            return false;
        }
        return Float.compare(this.score, other.score)==0;
    }

    /**
     * <p>Returns the hash code value for this object. The hash code is defined
     * by the following calculation:
     * </p>
     * <pre>
     *  int hash = 7;
     *  hash = 37 * hash + this.sampleIndex1;
     *  hash = 37 * hash + this.hap1;
     *  hash = 37 * hash + this.sampleIndex2;
     *  hash = 37 * hash + this.hap2;
     *  hash = 37 * hash + this.chromIndex;
     *  hash = 37 * hash + this.start;
     *  hash = 37 * hash + this.end;
     *  hash = 37 * hash + Float.floatToIntBits(this.score);
     </pre>
     * @return the hash code value for this object.
     */
    @Override
    public int hashCode() {
        int hash = 7;
        hash = 37 * hash + this.sampleIndex1;
        hash = 37 * hash + this.hap1;
        hash = 37 * hash + this.sampleIndex2;
        hash = 37 * hash + this.hap2;
        hash = 37 * hash + this.chromIndex;
        hash = 37 * hash + this.start;
        hash = 37 * hash + this.end;
        hash = 37 * hash + Float.floatToIntBits(this.score);
        return hash;
    }

    /**
     * <p>Compares this {@code StringIdSegment} with the specified
     * {@code StringIdSegment} for order.  Returns a negative integer,
     * zero, or positive integer as this object is less than, equal to,
     * or greater than the specified object.
     * </p>
     * {@code StringIdSegment} objects are ordered by comparing values of
     * {@code chromIndex()}, {@code start()}, {@code end()},
     * {@code score()} (using {@code Float.compare()}),
     * {@code sample1()}, {@code sample2()}, {@code hap1()},
     * and {@code hap2()} in that order.  All numerical values
     * are ordered by increasing magnitude, and all strings are ordered
     * by lexicographical order.
     *
     * @param segment the IBD segment to be compared with {@code this}
     * for order.
     *
     * @return a negative integer, zero, or positive integer as this object
     * is less than, equal to, or greater than the specified object.
     */
    @Override
    public int compareTo(StringIdSegment segment) {
        if (this.chromIndex != segment.chromIndex) {
            return (this.chromIndex < segment.chromIndex) ? -1 : 1;
        }
        if (this.start != segment.start) {
            return (this.start < segment.start) ? -1 : 1;
        }
        if (this.end != segment.end) {
            return (this.end < segment.end) ? -1 : 1;
        }
        int cmp = Float.compare(this.score, segment.score);
        if (cmp != 0) {
            return cmp;
        }
        if (this.sampleIndex1 != segment.sampleIndex1) {
            return (this.sampleIndex1 < segment.sampleIndex1) ? -1 : 1;
        }
        if (this.sampleIndex2 != segment.sampleIndex2) {
            return (this.sampleIndex2 < segment.sampleIndex2) ? -1 : 1;
        }
        if (this.hap1!=segment.hap1) {
            return (this.hap1 < segment.hap1) ? -1 : 1;
        }
        if (this.hap2!=segment.hap2) {
             return (this.hap2 < segment.hap2) ? -1 : 1;
        }
        return 0;
    }
}
